import numpy as np 
import matplotlib.pyplot as plt
import glob
import skimage
import pims
import skimage.transform
from tifffile import imsave
import sys

if len(sys.argv)>1:
	filename = sys.argv[1]
	print 'Stitching', filename
else:
	filename = '2015_12_04_V.glob_e3_concatenated.tif'

base_px_size = 0.48

px_sizes = np.loadtxt(filename[:-4]+'.txt')
px_sizes = {int(round(key)):v for key,v in px_sizes}

output_name = ".".join(filename.split('.')[:-1])+'_amira.tif'

other_filesnames = sorted(glob.glob(".".join(filename.split('.')[:-1])+'/*.tif'))
indeces = []
for fname in other_filesnames:
	f = fname.split('/')[-1]
	n = ''
	for s in f:
		if s.isdigit():
			n += s
		elif len(n)>0:
			break
	indeces.append(int(n))

base_imgs = pims.open(filename)

len_sequence =  max([max(px_sizes.keys()),len(base_imgs)])

output = np.zeros((len_sequence, base_imgs[0].shape[0], base_imgs[0].shape[1]), 'uint16')
for i, im in enumerate(base_imgs):
	output[i,:,:] = im

def transform(img, tx, ty, rotation):
	if abs(degree)<0.1:
		trans = skimage.transform.SimilarityTransform(translation=[trans_x, trans_y])
		transed = skimage.transform.warp(img, trans.inverse)
	else:
		shift_y, shift_x = np.array(img.shape[:2]) / 2.
		tf_rotate = skimage.transform.SimilarityTransform(rotation=np.deg2rad(degree))
		tf_shift = skimage.transform.SimilarityTransform(translation=[-shift_x, -shift_y])
		tf_shift_inv = skimage.transform.SimilarityTransform(translation=[shift_x, shift_y])
		trans = skimage.transform.SimilarityTransform(translation=[trans_x, trans_y])
		transed = skimage.transform.warp(img, (tf_shift + (tf_rotate + tf_shift_inv) + trans).inverse)
	return transed

prev_defined = False
for i, img_i in enumerate(indeces):
	print i+1, '/', len(indeces)
	base_i = img_i - 1

	if prev_defined and prev_index == base_i-1:
		base = previous/65536.
	else:
		base = base_imgs[base_i]/65536.
	
	img = pims.open(other_filesnames[i])[0][::-1,::-1]
	img = skimage.transform.rescale(img, px_sizes[img_i]/base_px_size)

	# Match size
	if img.shape[0] < base.shape[0]:
		img2 = np.zeros((base.shape[0],img.shape[1]))
		img2[:img.shape[0],:img.shape[1]] = img
		img = img2
	if img.shape[1] < base.shape[1]:
		img2 = np.zeros((img.shape[0], base.shape[1]))
		img2[:img.shape[0],:img.shape[1]] = img
		img = img2
	orig_img = img.copy()

	print 'Base transformation'
	best = float('infinity')
	for trans_x in range(-250,250,10):
		for trans_y in range(-250,250,10):
			for degree in [0,90,180,270]:
				transed = transform(img, trans_x, trans_y, degree)
				transed = transed[:base.shape[0],:base.shape[1]]

				err = np.sum((base-transed)**2)
				if err < best:
					best = err
					best_trans = transed
					best_stats = (trans_x, trans_y, degree)
	print '    Base transformation:', best_stats, 'with score', best
	first_transform = best_stats

	img = best_trans
	print 'Refining'
	best = float('infinity')
	for trans_x in range(-10,10,2):
		for trans_y in range(-10,10,2):
			degree = 0

			transed = transform(img, trans_x, trans_y, degree)
			transed = transed[:base.shape[0],:base.shape[1]]

			err = np.sum((base-transed)**2)
			if err < best:
				best = err
				best_trans = transed
				best_stats = (trans_x, trans_y, degree)

	print '   Extra correction:', best_stats, 'with score', best

	
	transed = 65536 * best_trans

	output[base_i,:,:] = np.array(transed, dtype=np.uint16)

	previous = transed.copy()
	prev_defined = True
	prev_index = base_i

	# base = base_imgs[base_i]
	# plt.subplot(1,2,1)
	# plt.imshow(base)
	# plt.subplot(1,2,2)
	# plt.imshow(transed)
	# plt.figure()
	# plt.imshow(base + transed)
	# plt.show()


output2 = np.zeros((len_sequence, base_imgs[0].shape[1], base_imgs[0].shape[0]), 'uint16')
for i in xrange(len_sequence):
	output2[i,:,:] = output[i,::-1,:].T

imsave(output_name, output2)
