##################### SETTINGS #####################
N = 100
####################################################

import matplotlib
matplotlib.use('TkAgg')
from numpy import arange, sin, pi
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg
# implement the default mpl key bindings
from matplotlib.backend_bases import key_press_handler
from matplotlib.figure import Figure
import sys
if sys.version_info[0] < 3:
	import Tkinter as Tk
else:
	import tkinter as Tk
import tkMessageBox
import tkFileDialog
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 
from scipy import interpolate,interp,ndimage
from multiprocessing import Pool, cpu_count, Lock
import pims
import json
import time
import os
import cPickle
import scipy
from anisodiff import anisodiff, bandpass
from shutil import copyfile
import errno
import colormaps
import skimage
import skimage.morphology
from copy import copy

windows = os.name=='nt'
def silentremove(filename):
    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise

point_mode_str = 'Initialisation mode'
correction_mode_str = 'Correction mode'
snakes_has_been_run = False
parallel = True

def poolinit(l):
    global lock
    lock = l

if __name__=='__main__':
	######### SET UP FIGURE ########
	root = Tk.Tk() 
	root.wm_title('Volvox Inversion Tracker')
	fig = Figure(figsize=(7,7), dpi=100)
	title = fig.suptitle(point_mode_str)
	ax = fig.add_subplot(111)

	canvas = FigureCanvasTkAgg(fig, master=root)
	canvas.show()
	canvas.get_tk_widget().pack(side=Tk.TOP, fill=Tk.BOTH, expand=1)

	toolbar = NavigationToolbar2TkAgg( canvas, root )
	toolbar.update()
	canvas._tkcanvas.pack(side=Tk.TOP, fill=Tk.BOTH, expand=1)

	fullfilename = tkFileDialog.askopenfilename() # 'imgs/Embryo6_Center.tif'
	#fullfilename = 'imgs/2015_12_02_V.glob_e3.tif'
else:
	with open('comm/cluster.temp') as f:
		fullfilename = f.read()

imgs = pims.open(fullfilename)
filename = "".join(fullfilename.split('/')[-1].split('.')[:-1]).replace(' ','_')
print filename

if __name__=='__main__':
	point_mode = True

	current_index = 0
	print 'trackerdata/points_%s.p'%filename
	if os.path.exists('trackerdata/points_%s.p'%filename):
		f = open("trackerdata/points_%s.p"%filename)
		x,y = cPickle.load(f)
		if len(x) < len(imgs):
			x += [list() for _ in range(len(imgs)-len(x))]
			y += [list() for _ in range(len(imgs)-len(y))]
		f.close()
	else:
		x = [list() for _ in range(len(imgs))]
		y = [list() for _ in range(len(imgs))]

	if os.path.exists('trackerdata/snakes_%s.p'%filename):
		snakes_has_been_run = True
		f = open("trackerdata/snakes_%s.p"%filename)
		print 'Snakes from previous session has been loaded.'
		print 'Use fixer!'
	##################################

	if not os.path.exists('trackerdata/'):
		os.mkdir('trackerdata')
	if not os.path.exists('comm/'):
		os.mkdir('comm')

def save():
	f = open("trackerdata/points_%s.p"%filename,"w")
	cPickle.dump((x,y),f)
	f.close()

def initial_spline(X,Y,n=20):
	tck,u = interpolate.splprep([X,Y],k=3,s=1)
	xx,yy = interpolate.splev(np.linspace(0,1,n),tck)
	return xx,yy

def interpolate_curve(current_index):
	i = current_index
	while i>=0 and len(x[i])<=3: i -= 1
	j = current_index
	while j<len(imgs) and len(x[j])<=3: j += 1

	if i>=0:
		xx1,yy1 = initial_spline(x[i],y[i],n=N)
	if j<len(imgs):
		xx2,yy2 = initial_spline(x[j],y[j],n=N)

	# If not inside range, return edge:
	if i<0:
		return xx2, yy2
	elif j>=len(imgs):
		return xx1, yy1

	# Interpolate otherwise:
	xx = (xx1*abs(current_index-j)+xx2*abs(current_index-i))/(abs(current_index-j)+abs(current_index-i))
	yy = (yy1*abs(current_index-j)+yy2*abs(current_index-i))/(abs(current_index-j)+abs(current_index-i))

	return xx, yy

def do_plot(t,overwrite=False):
	ax.clear()
	cmap = plt.get_cmap('Dark2')
	img = np.array(imgs[t],dtype=float)
	imgplot = ax.imshow(img) # , cmap=cm.Greys_r, vmin=0, vmax=255)
	ax.plot(x[current_index],y[current_index],'xr')
	if len(x[current_index])>3:
		xx,yy = initial_spline(x[current_index],y[current_index],n=N)
		ax.plot(xx,yy,'--k')
	elif len(x[current_index])==0 and len([1 for e in x if len(e)>3])>=1:
		xx,yy = interpolate_curve(current_index)
		ax.plot(xx,yy,'--k')

	ax.axis([0,img.shape[1],img.shape[0],0])
	canvas.draw()

def on_key_event(event):
	global current_index
	if event.key=='control':
		del x[current_index][-1]
		del y[current_index][-1]
		do_plot(current_index)
		save()

	if event.key=='right':
		prev = current_index
		current_index = current_index+1
		current_index = min([len(imgs)-1,current_index])
		print 'Frame =', current_index+1, '/', len(imgs)
		if prev!=current_index:
			do_plot(current_index)

	if event.key=='up':
		prev = current_index
		current_index = current_index+5
		current_index = min([len(imgs)-1,current_index])
		print 'Frame =', current_index+1, '/', len(imgs)
		if prev!=current_index:
			do_plot(current_index)

	if event.key=='left':
		prev = current_index
		current_index = current_index-1
		current_index = max([0,current_index])
		print 'Frame =', current_index+1, '/', len(imgs)
		if prev!=current_index:
			do_plot(current_index)

	if event.key=='down':
		prev = current_index
		current_index = current_index-5
		current_index = max([0,current_index])
		print 'Frame =', current_index+1, '/', len(imgs)
		if prev!=current_index:
			do_plot(current_index)
	key_press_handler(event, canvas, toolbar)

def on_mouse_event(event):
	global axes, choano, locating
	nx,ny = event.xdata, event.ydata
	if nx!=None:
		if event.button == 1:
			if (int(nx),int(ny)) not in zip(map(int,x[current_index]),
						map(int,y[current_index])):
				x[current_index].append(nx)
				y[current_index].append(ny)
				do_plot(current_index)
				canvas.draw()
				save()
		elif event.button == 3:
			xx = np.array(x[current_index])
			yy = np.array(y[current_index])
			if len(xx)>0:
				dist2 = (xx-nx)**2 + (yy-ny)**2
				i = np.argmin(dist2)
				del x[current_index][i]
				del y[current_index][i]
				do_plot(current_index)
				canvas.draw()
				save()
if __name__=='__main__':
	canvas.mpl_connect('key_press_event', on_key_event)
	canvas.mpl_connect('button_press_event', on_mouse_event)

####################################################

def compile():
	if windows:
		a = os.system('g++ -std=c++11 -o comm/%s.exe snake_it.cpp'%filename)
	else:
		a = os.system('g++ -std=c++11 -o comm/%s.out snake_it.cpp'%filename)
		print 'g++ -std=c++11 -o comm/%s.out snake_it.cpp'%filename
	if a!=0:
		raise ValueError("Could not compile.")

def snake_all():
	try:
		os.mkdir('track')
	except OSError:
		pass
	args = []
	init_snakes = []
	for index in xrange(len(imgs)):
		if len(x[index])>3:
			xx,yy = initial_spline(x[index],
						y[index],n=N)
		elif len(x[index])==0:
			xx,yy = interpolate_curve(index)
		args.append((xx,yy,index))
		init_snakes.append(dict())
		init_snakes[-1]['x'] = xx
		init_snakes[-1]['y'] = yy

	print 'Using %i CPUs'%cpu_count()
	with open('comm/cluster.temp','w') as f:
		f.write(fullfilename)

	if parallel:
		l = Lock()
		pool = Pool(cpu_count(),initializer=poolinit, initargs=(l,))
		snakes = pool.map(snake_one,args,chunksize=1)
	else:
		snakes = map(snake_one,args)
	
	for i,s in enumerate(snakes):
		if s == False or np.isnan(s['y']).any() or np.isnan(s['x']).any():
			print 'Interpolating frame', i+1
			i1 = next((j for j in range(i-1,-1,-1) if (snakes[j]!=False
							and not np.isnan(snakes[j]['y']).any() or not np.isnan(snakes[j]['x']).any())))
			try:
				i2 = next((j for j in range(i+1,len(snakes)) if (snakes[j]!=False
							and not np.isnan(snakes[j]['y']).any() or not np.isnan(snakes[j]['x']).any())))
			except StopIteration:
				for j in xrange(i,len(snakes)):
					del snakes[i]
				break
			snakes[i] = {}
			snakes[i]['x'] = [interp(i,[i1,i2],[snakes[i1]['x'][k],snakes[i2]['x'][k]])
									for k in xrange(len(snakes[i1]['x']))]
			snakes[i]['y'] = [interp(i,[i1,i2],[snakes[i1]['y'][k],snakes[i2]['y'][k]])
									for k in xrange(len(snakes[i1]['y']))]
	
	with open("trackerdata/snakes_%s.p"%filename,'w') as f:
		cPickle.dump(snakes,f)

	print 'Dumping init snakes'
	with open("trackerdata/init_snakes_%s.p"%filename,'w') as f:
		print 'Dumped.'
		cPickle.dump(init_snakes,f)

	silentremove('comm/cluster.temp')
	print 'Finished!'

def cleanup():
	silentremove('comm/cluster.temp')
	if windows:
		silentremove('comm/%s.exe'%filename)
		silenremove('comm/%s.exe'%this_filename)
	else:
		silentremove('comm/%s.out'%filename)
		silentremove('comm/%s.out'%this_filename)
	for index in xrange(len(imgs)):
		this_filename = 'comm/'+filename+str(index)
		silentremove(this_filename+'_trans.temp')
		silentremove(this_filename+'_trans_out.temp')

def energy(x,y,init_x,init_y,close_x,close_y,img,pref_dist=1.,a=0,b=0,g=1.,ki=1.,kc=1.):
	try:
		dist = np.mean(np.sqrt((x[:-1]-x[1:])**2+(y[:-1]-y[1:])**2))

		curvatures = np.sqrt((2*x[1:-1]-x[:-2]-x[2:])**2+(2*y[1:-1]-y[:-2]-y[2:])**2)
		curvature = np.mean(list(curvatures[curvatures>1])+[0])

		intensity = np.mean(img[np.round(y).astype(np.int32),np.round(x).astype(np.int32)])

		close_moved = np.mean(np.sqrt(  np.min((x[:,None]-close_x[None,:])**2 + (y[:,None]-close_y[None,:])**2,1)  ))
		init_moved = np.mean(np.sqrt((x-init_x)**2 + (y-init_y)**2))

		return a * (dist-pref_dist)**2 + b * curvature - g * intensity + ki * init_moved + kc * close_moved
	except IndexError:
		return 1e12
		
def snake_one(args):
	init_x,init_y,index = args

	print 'Snaking frame '+str(index+1)
	iterations = 10000
	a = 0.1
	b = 0.5
	g = 1.
	after_g = 5.
	after_ki = 0.5
	ki = 0.1
	kc = 2.
	
	init_x = np.array(init_x)
	init_y = np.array(init_y)
	pref_dist = np.mean(np.sqrt((init_x[:-1]-init_x[1:])**2+(init_y[:-1]-init_y[1:])**2))

	if parallel:
		lock.acquire()
		img = np.array(imgs[index], dtype=np.float32)
		lock.release()
	else:
		img = np.array(imgs[index], dtype=np.float32)

	if np.max(img)>1.1:
		if np.max(img)>256:
			img = skimage.exposure.rescale_intensity(img, in_range=(0, 2**14), out_range=(0,1.))
		else:
			img = skimage.exposure.rescale_intensity(img, in_range=(0, 2**8), out_range=(0,1.))

	blur = bandpass(img,0.5,50)
	#blur = ndimage.median_filter(blur, 1)
	blur /= (0.01 + ndimage.gaussian_filter(blur, sigma=30))
	blur = ndimage.gaussian_filter(blur, 3)


	skel = np.mean([skimage.morphology.skeletonize(blur>x) for x in [0.3,0.4,0.5,0.6,0.7]],0)
	skelblur = ndimage.gaussian_filter(skel, 3)
	skel = skimage.morphology.skeletonize(skelblur>0.07)

	skel_y,skel_x = np.where(skel)

	x = np.array(init_x)
	y = np.array(init_y)

	idx = np.arange(N,dtype=np.float32)
	E0 = energy(x,y,init_x,init_y,skel_x,skel_y,blur,pref_dist=pref_dist,a=a,b=b,g=g,ki=ki,kc=kc)
	for ite in xrange(iterations):
		if ite>iterations/2 and kc>0:
			kc = 0
			g = after_g
			ki = after_ki
			E0 = energy(x,y,init_x,init_y,skel_x,skel_y,blur,pref_dist=pref_dist,a=a,b=b,g=g,ki=ki,kc=kc)

		# Random move:
		i = np.random.randint(N)
		sigma = 1 + np.abs(np.random.normal(1,3))
		weights = np.exp(-((idx-i)/sigma)**2)
		dr = 1 + np.abs(np.random.normal(1,5))
		theta = 2*np.pi*np.random.random()
		dx = weights*dr*np.cos(theta)
		dy = weights*dr*np.sin(theta)
		x[1:-1] += dx[1:-1]
		y[1:-1] += dy[1:-1]
		E = energy(x,y,init_x,init_y,skel_x,skel_y,blur,pref_dist=pref_dist,a=a,b=b,g=g,ki=ki,kc=kc)
		if E<E0:
			E0 = E
		else:
			x[1:-1] -= dx[1:-1]
			y[1:-1] -= dy[1:-1]

	# ax.clear()
	# ax.imshow(blur,cmap=colormaps.viridis)
	# ax.plot(init_x,init_y,'-b')
	# ax.plot(skel_x,skel_y,'.k')
	# ax.plot(x,y,'-xr')
	# canvas.draw()

	d = {"x":list(x),"y":list(y)}
	return d

# ################# BUTTONS ############### 
if __name__=='__main__':
	
	def _quit():
		root.quit()	 # stops mainloop

		root.destroy()  # this is necessary on Windows to prevent
						# Fatal Python Error: PyEval_RestoreThread: NULL tstate
	btn = Tk.Button(master=root, text='Quit', command=_quit)
	btn.pack(side=Tk.RIGHT)

	def _deletepoints():
		result = tkMessageBox.askquestion("Delete current frame points", "Are you sure?", icon='warning')
		if result == 'yes':
			x[current_index] = []
			y[current_index] = [] 
			do_plot(current_index)
			save()
	btn = Tk.Button(master=root, text='Delete all points in this frame',
				 command=_deletepoints)
	btn.pack(side=Tk.RIGHT)

	def _delete_all_frames():
		result = tkMessageBox.askquestion("Delete all points in all frames", "Are you sure?", icon='warning')
		if result == 'yes':
			for i in xrange(len(x)):
				x[i] = []
				y[i] = []
			do_plot(current_index)
			save()
	btn = Tk.Button(master=root, text='Delete all points in all frames',
				 command=_delete_all_frames)
	btn.pack(side=Tk.RIGHT)

	def _starttrack():
		print 'Starting track using %i manual initialisations...'%len([1 for e in x if len(e)>3])
		snake_all()
	btn = Tk.Button(master=root, text='Start track', command=_starttrack)
	btn.pack(side=Tk.RIGHT)

	import time
	########### MAIN LOOP ##############
	# for i in xrange(45,70):
	# 	print i
	# 	current_index = i
	# 	do_plot(current_index)
	# 	time.sleep(1)
	#snake_all()
	do_plot(current_index)
	#
	Tk.mainloop()
