##################### SETTINGS #####################
N = 100
####################################################

import matplotlib
matplotlib.use('TkAgg')
from numpy import arange, sin, pi
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg
# implement the default mpl key bindings
from matplotlib.backend_bases import key_press_handler
from matplotlib.figure import Figure
import sys
if sys.version_info[0] < 3:
	import Tkinter as Tk
else:
	import tkinter as Tk
import tkMessageBox
import tkFileDialog
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 
from scipy import interpolate,interp,ndimage
from multiprocessing import Pool, cpu_count
import pims
import json
import time
import os
import cPickle
from copy import deepcopy
from shutil import copyfile
import errno
from anisodiff import anisodiff, bandpass
windows = os.name=='nt'
def silentremove(filename):
    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise

point_mode_str = 'Initialisation mode'
correction_mode_str = 'Correction mode'
snakes_has_been_run = False

def line_segment_distance(v,w,p):
	# line segment from v to w. Returns distance to p.
	l2 = np.sum((v-w)**2)
	if l2 == 0:
		return np.sqrt(np.sum((v-p)**2))
	t = np.dot(p-v,w-v)/l2;
	if t<0:
		return np.sqrt(np.sum((v-p)**2))
	if t>1:
		return np.sqrt(np.sum((w-p)**2))
	p_line = v + t *(w-v)
	return np.sqrt(np.sum((p_line-p)**2))

if  __name__=='__main__':
	######### SET UP FIGURE ########
	root = Tk.Tk() 
	root.wm_title('Volvox Inversion Tracker')
	fig = Figure(figsize=(7,7), dpi=100)
	title = fig.suptitle('Fixer')
	ax = fig.add_subplot(111)

	canvas = FigureCanvasTkAgg(fig, master=root)
	canvas.show()
	canvas.get_tk_widget().pack(side=Tk.TOP, fill=Tk.BOTH, expand=1)

	toolbar = NavigationToolbar2TkAgg( canvas, root )
	toolbar.update()
	canvas._tkcanvas.pack(side=Tk.TOP, fill=Tk.BOTH, expand=1)

	fullfilename = tkFileDialog.askopenfilename()
	#fullfilename = 'imgs/2015_11_30_V.glob_e1_concatenated.tif'

	with open('comm/cluster.temp','w') as f:
		f.write(fullfilename)
else:
	with open('comm/cluster.temp') as f:
		fullfilename = f.read()

imgs = pims.open(fullfilename)
filename = "".join(fullfilename.split('/')[-1].split('.')[:-1]).replace(' ','_')

def initial_spline(X,Y,n=50):
	while True:
		X = np.array(X,dtype=np.float64)
		Y = np.array(Y,dtype=np.float64)
		dist2 = (X[None,:]-X[:,None])**2 \
			+ (Y[None,:]-Y[:,None])**2
		idx = np.argwhere(np.tril(dist2<1,k=-1))
		if len(idx)==0:
			break
		else:
			X = list(X)
			Y = list(Y)
			del X[idx[0][0]]
			del Y[idx[0][0]]
	tck,u = interpolate.splprep([X,Y],k=1,s=0.5)
	xx,yy = interpolate.splev(np.linspace(0,1,n),tck)
	return xx,yy

if True or __name__=='__main__':
	point_mode = True
	current_index = 0
	if os.path.exists('trackerdata/snakes_%s.p'%filename):
		print 'trackerdata/snakes_%s.p'%filename
		snakes_has_been_run = True
		f = open("trackerdata/snakes_%s.p"%filename)
		l = cPickle.load(f)
		sx = list(); sy = list()
		orig_sx = list(); orig_sy = list()

		x,y = initial_spline(l[0]['x'],l[0]['y'],n=50)
		for i,d in enumerate(l):
			x,y = initial_spline(d['x'],d['y'],n=50)
			sx.append(list(x))
			sy.append(list(y))
			orig_sx.append(list(x))
			orig_sy.append(list(y))
		f.close()
	else:
		raise ValueError('No tracing found')

	if os.path.exists('trackerdata/init_snakes_%s.p'%filename):
		print 'trackerdata/init_snakes_%s.p'%filename
		init_snakes_found = True
		with open("trackerdata/init_snakes_%s.p"%filename) as f:
			l = cPickle.load(f)
		init_sx = list(); init_sy = list()
		for i,d in enumerate(l):
			x,y = initial_spline(d['x'],d['y'],n=50)
			init_sx.append(list(x))
			init_sy.append(list(y))
	else:
		init_snakes_found = False

	##################################

def save():
	f = open("trackerdata/snakes_%s.p"%filename,'w')
	snakes = [{'x':sx[i],'y':sy[i]} for i in xrange(len(sx))]
	cPickle.dump(snakes,f)
	f.close()

def do_plot(t,overwrite=False):
	ax.clear()
	cmap = plt.get_cmap('Dark2')
	img = np.array(imgs[t],dtype=float)
	imgplot = ax.imshow(img) # , cmap=cm.Greys_r, vmin=0, vmax=255)
	ax.plot(sx[current_index],sy[current_index],'.-k')
	#print sx[current_index]

	ax.axis([0,img.shape[1],img.shape[0],0])
	canvas.draw()

def on_key_event(event):
	global current_index, sx, sy

	number = False
	try:
		number = int(event.key) + 1
	except:
		pass

	# if event.key=='control':
	# 	del x[current_index][-1]
	# 	del y[current_index][-1]
	# 	do_plot(current_index)
	# 	save()

	if event.key=='right':
		prev = current_index
		current_index = current_index+1
		current_index = min([len(imgs)-1,current_index])
		print 'Frame =', current_index+1, '/', len(imgs)
		if prev!=current_index:
			do_plot(current_index)

	if event.key=='up':
		prev = current_index
		current_index = current_index+5
		current_index = min([len(imgs)-1,current_index])
		print 'Frame =', current_index+1, '/', len(imgs)
		if prev!=current_index:
			do_plot(current_index)

	if event.key=='left':
		prev = current_index
		current_index = current_index-1
		current_index = max([0,current_index])
		print 'Frame =', current_index+1, '/', len(imgs)
		if prev!=current_index:
			do_plot(current_index)

	if event.key=='down':
		prev = current_index
		current_index = current_index-5
		current_index = max([0,current_index])
		print 'Frame =', current_index+1, '/', len(imgs)
		if prev!=current_index:
			do_plot(current_index)

	if number and event.xdata!=None:
		radius2 = (10*number)**2
		nx = event.xdata
		ny = event.ydata

		idxs = np.argwhere(((np.array(sx[current_index])-nx)**2
						+ (np.array(sy[current_index])-ny)**2) < radius2).ravel()
		idxs.sort()
		if len(idxs)>0: # remove points within radius
			if np.max(idxs[1:]-idxs[:-1])>5:
				swap = idxs[np.argwhere((idxs[1:]-idxs[:-1])>5).ravel()[0]]
				to_be_removed = sorted(range(idxs[0],idxs[swap]+1) + range(idxs[swap+1],idxs[-1]+1))
				to_be_removed = list(reversed(to_be_removed))
			else:
				to_be_removed = list(reversed(range(np.min(idxs),np.max(idxs)+1)))

			for tbr in to_be_removed:
				del sx[current_index][tbr]
				del sy[current_index][tbr]

			do_plot(current_index)


		idxs = np.argwhere(((np.array(init_sx[current_index])-nx)**2
						+ (np.array(init_sy[current_index])-ny)**2) < radius2).ravel()
		idxs.sort()
		if len(idxs)>0: # introduce init points
			if np.max(idxs[1:]-idxs[:-1])>5:
				swap = idxs[np.argwhere((idxs[1:]-idxs[:-1])>5).ravel()[0]]
				to_be_introduced_list = [sorted(range(idxs[0],idxs[swap]+1)),
									     sorted(range(idxs[swap+1],idxs[-1]+1))]
			else:
				to_be_introduced_list = [range(np.min(idxs),np.max(idxs)+1)]

			print to_be_introduced_list
			for to_be_introduced in to_be_introduced_list:
				tbr = to_be_introduced[0]

				xx = np.array(sx[current_index])
				yy = np.array(sy[current_index])
				dist2 = (xx-init_sx[current_index][tbr])**2 + (yy-init_sy[current_index][tbr])**2
				i = np.argmin(dist2)

				p = np.array([init_sx[current_index][tbr],init_sy[current_index][tbr]])
				x0 = np.array([xx[i],yy[i]])
				if i==0:
					x1 = np.array([3*xx[i]-2*xx[i+1],3*yy[i]-2*yy[i+1]])
					x2 = np.array([xx[i+1],yy[i+1]])
					if line_segment_distance(x0,x1,p) < line_segment_distance(x0,x2,p):
						insert_at = 0
					else:
						insert_at = 1
				elif i==len(xx)-1:
					x1 = np.array([xx[i-1],yy[i-1]])
					x2 = np.array([3*xx[i]-2*xx[i-1],3*yy[i]-2*yy[i-1]])
					if line_segment_distance(x0,x1,p) < line_segment_distance(x0,x2,p):
						insert_at = len(xx)-1
					else:
						insert_at = len(xx)
				else:
					x1 = np.array([xx[i-1],yy[i-1]])
					x2 = np.array([xx[i+1],yy[i+1]])
					if line_segment_distance(x0,x1,p) < line_segment_distance(x0,x2,p):
						insert_at = i
					else:
						insert_at = i+1

				for tbr in to_be_introduced:
					print tbr
					sx[current_index].insert(insert_at,init_sx[current_index][tbr])
					sy[current_index].insert(insert_at,init_sy[current_index][tbr])
					insert_at += 1

			do_plot(current_index)
		save()

	key_press_handler(event, canvas, toolbar)

def on_mouse_event(event):
	global axes, choano, locating, sx, sy
	nx,ny = event.xdata, event.ydata
	if nx!=None:
		xx = np.array(sx[current_index])
		yy = np.array(sy[current_index])
		dist2 = (xx-nx)**2 + (yy-ny)**2
		i = np.argmin(dist2)

		if event.button == 3:
			del sx[current_index][i]
			del sy[current_index][i]
		elif event.button == 1:
			# Find insert_at = index to inset before
			p = np.array([nx,ny])
			x0 = np.array([xx[i],yy[i]])

			if i==0:
				x1 = np.array([3*xx[i]-2*xx[i+1],3*yy[i]-2*yy[i+1]])
				x2 = np.array([xx[i+1],yy[i+1]])
				if line_segment_distance(x0,x1,p) < line_segment_distance(x0,x2,p):
					insert_at = 0
				else:
					insert_at = 1
			elif i==len(xx)-1:
				x1 = np.array([xx[i-1],yy[i-1]])
				x2 = np.array([3*xx[i]-2*xx[i-1],3*yy[i]-2*yy[i-1]])
				if line_segment_distance(x0,x1,p) < line_segment_distance(x0,x2,p):
					insert_at = len(xx)-1
				else:
					insert_at = len(xx)
			else:
				x1 = np.array([xx[i-1],yy[i-1]])
				x2 = np.array([xx[i+1],yy[i+1]])
				if line_segment_distance(x0,x1,p) < line_segment_distance(x0,x2,p):
					insert_at = i
				else:
					insert_at = i+1

			sx[current_index].insert(insert_at,nx)
			sy[current_index].insert(insert_at,ny)
		elif event.button == 2 and init_snakes_found:
			
			if dist2[i] < 5**2:
				del sx[current_index][i]
				del sy[current_index][i]
				if i == len(sx[current_index]):
					i -= 1

			ixx = np.array(init_sx[current_index])
			iyy = np.array(init_sy[current_index])
			dist2 = (ixx-nx)**2 + (iyy-ny)**2
			j = np.argmin(dist2)
			p = np.array([ixx[j],iyy[j]])

			tries = [iii for iii in np.arange(-1,3) + i if 0<=iii<len(xx)]
			min_dist = float('infinity')
			insert_at = 0
			for index2 in tries:
				csx = deepcopy(sx[current_index])
				csy = deepcopy(sy[current_index])
				csx.insert(index2,p[0])
				csy.insert(index2,p[1])
				csx = np.array(csx)
				csy = np.array(csy)
				dist2 = np.sum((csx[1:]-csx[:-1])**2 + (csy[1:]-csy[:-1])**2)
				if min_dist > dist2:
					min_dist = dist2
					insert_at = index2
			####

			sx[current_index].insert(insert_at,p[0])
			sy[current_index].insert(insert_at,p[1])

	do_plot(current_index)
	save() # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

def export():
	f = open("track/%s.txt"%filename,'w')
	f.write('t    x    y\n')
	for t in xrange(len(sx)):
		x,y = initial_spline(sx[t],sy[t],n=50)
		if not np.isnan(x).any():
			for i in xrange(len(x)):
				f.write('%i    %i    %i\n'%(t,x[i],y[i]))
		else:
			print 'Failed export at', t
	f.close()
	print 'Exported to', "track/%s.txt"%filename

def export_img(t):
	print 'Printing '+str(t)
	x,y = initial_spline(sx[t],sy[t],n=50)
	plt.clf()
	img = np.array(imgs[t],dtype=float)
	plt.imshow(img,cmap=cm.jet)
	plt.plot(x,y,'-xk')
	plt.axis([0,img.shape[1],img.shape[0],0])
	plt.savefig('track/%s_final/%05i.png'%(filename,t))

def export_imgs():
	try:
		os.mkdir('track/%s_final/'%filename)
	except OSError:
		pass
	map(export_img,range(len(sx))) # do not try and parallelize this
								   # for some reason it takes all RAM
	print 'All exported to', 'track/%s_final/'%filename

if __name__=='__main__':
	canvas.mpl_connect('key_press_event', on_key_event)
	canvas.mpl_connect('button_press_event', on_mouse_event)

# ################# BUTTONS ############### 
if __name__=='__main__':
	def _quit():
		root.quit()	 # stops mainloop

		root.destroy()  # this is necessary on Windows to prevent
						# Fatal Python Error: PyEval_RestoreThread: NULL tstate
	btn = Tk.Button(master=root, text='Quit', command=_quit)
	btn.pack(side=Tk.RIGHT)

	btn = Tk.Button(master=root, text='Export images', command=export_imgs)
	btn.pack(side=Tk.RIGHT)

	btn = Tk.Button(master=root, text='Export', command=export)
	btn.pack(side=Tk.RIGHT)

	########### MAIN LOOP ##############
	do_plot(current_index)
	Tk.mainloop()
